from unittest.mock import Mock

import numpy as np

from choix.convergence import NormOfDifferenceTest, ScalarFunctionTest


def test_sf_works():
    """Basic test for ScalarFunctionTest."""
    params = np.arange(5, dtype=float)
    seq = [2.0, 1.0, 2.0, 1.5, 1.3]
    fun = Mock()
    # With `tol=0.7`.
    fun.side_effect = seq
    test = ScalarFunctionTest(fun, tol=0.7)
    assert not test(params)
    assert not test(params)
    assert not test(params)
    assert test(params)
    assert test(params)
    # With `tol=0.5`.
    fun.side_effect = seq
    test = ScalarFunctionTest(fun, tol=0.5)
    assert not test(params)
    assert not test(params)
    assert not test(params)
    assert not test(params)
    assert test(params)


def test_nod_works():
    """Basic test for NormOfDifferenceTest."""
    # With `tol=0.7`.
    test = NormOfDifferenceTest(tol=0.7, order=1)
    assert not test(np.exp([-0.5, +0.5]))
    assert not test(np.exp([+0.5, -0.5]))
    assert test(np.exp([0.0, 0.0]))
    assert test(np.exp([-0.2, +0.2]))
    # With `tol=0.4`.
    test = NormOfDifferenceTest(tol=0.4, order=1)
    assert not test(np.exp([-0.5, +0.5]))
    assert not test(np.exp([+0.5, -0.5]))
    assert not test(np.exp([0.0, 0.0]))
    assert test(np.exp([-0.2, +0.2]))


def test_nod_norm():
    """NODTest should respect the order of the norm."""
    # With L1 norm.
    test = NormOfDifferenceTest(tol=0.8, order=1)
    assert not test(np.array([-0.5, +0.5]))
    assert not test(np.array([+0.5, -0.5]))
    # With L2 norm.
    test = NormOfDifferenceTest(tol=0.8, order=2)
    assert not test(np.array([-0.5, +0.5]))
    assert test(np.array([+0.5, -0.5]))


def test_sf_no_update():
    """SFTest should not update its state when `update=False`."""
    fun = Mock(side_effect=[2.0, 1.2, 0.9, 0.8])
    test = ScalarFunctionTest(fun, tol=1.0)
    dummy = np.arange(5, dtype=float)
    assert not test(dummy)
    # `2.0 - 1.2 < 1.0` -> True.
    assert test(dummy, update=False)
    # `2.0 - 0.9 > 1.0` -> False.
    assert not test(dummy)
    # `0.9 - 0.8 < 1.0` -> True
    assert test(dummy)


def test_nod_no_update():
    """NODTest should not update its state when `update=False`."""
    test = NormOfDifferenceTest(tol=0.6, order=1)
    assert not test(np.array([-1.0, +1.0]))
    # Average norm of difference = 0.5.
    assert test(np.array([-0.5, +0.5]), update=False)
    # Average norm of difference = 0.8.
    assert not test(np.array([-0.2, +0.2]))
    # Average norm of difference = 0.1.
    assert test(np.array([-0.1, +0.1]))


def test_nod_normalization():
    """NODTest should normalize the criterion with the # of items."""
    for i in (1, 2, 5, 10):
        test = NormOfDifferenceTest(tol=0.5, order=1)
        assert not test(np.array([-1.7, 1.7] * i))
        assert not test(np.array([-1.0, 1.0] * i))
        assert test(np.array([-0.8, 0.8] * i))



from math import e, log

import networkx as nx
import numpy as np
import pytest
import scipy.stats as sps
from scipy.linalg import inv

from choix.utils import (
    compare,
    footrule_dist,
    generate_pairwise,
    generate_params,
    generate_rankings,
    inv_posdef,
    kendalltau_dist,
    log_likelihood_network,
    log_likelihood_pairwise,
    log_likelihood_rankings,
    log_likelihood_top1,
    normal_cdf,
    normal_pdf,
    probabilities,
    rmse,
    softmax,
    statdist,
)

RND = np.random.RandomState(42)


def test_footrule_dist_error():
    params1 = np.arange(3, dtype=float)
    params2 = np.arange(4, dtype=float)
    with pytest.raises(AssertionError):
        footrule_dist(params1, params2)


def test_footrule_dist_simple_cases():
    params1 = np.array([+1.0, -1.2, +0.0])
    params2 = np.array([+1.5, -0.2, -0.2])
    params3 = np.array([-1.0, +1.2, +0.0])
    for params in (params1, params2, params3):
        assert footrule_dist(params, params) == 0.0
    assert footrule_dist(params1, params2) == 1.0
    assert footrule_dist(params1, params3) == 4.0


def test_footrule_dist_default():
    params1 = np.arange(0, 10, dtype=float)
    assert footrule_dist(params1) == 0
    params2 = np.arange(0, -10, -1, dtype=float)
    assert footrule_dist(params2) == 10**2 / 2
    params3 = np.ones(10, dtype=float)
    assert footrule_dist(params3) == 25.0


def test_kendalltau_dist_error():
    params1 = np.arange(3, dtype=float)
    params2 = np.arange(4, dtype=float)
    with pytest.raises(AssertionError):
        kendalltau_dist(params1, params2)


def test_kendalltau_dist_simple_cases():
    params1 = np.array([+1.0, -1.2, +0.0])
    params2 = np.array([+1.5, -0.2, -0.2])
    params3 = np.array([-1.0, +1.2, +0.0])
    for params in (params1, params2, params3):
        assert kendalltau_dist(params, params) == 0.0
    # Tie broken in a "fortunate" way.
    assert kendalltau_dist(params1, params2) == 0.0
    assert kendalltau_dist(params1, params3) == 3.0
    assert kendalltau_dist(params2, params3) == 3.0


def test_kendalltau_dist_default():
    params1 = np.arange(0, 10, dtype=float)
    assert kendalltau_dist(params1) == 0
    params2 = np.arange(0, -10, -1, dtype=float)
    assert kendalltau_dist(params2) == (10 * 9) / 2
    # This is a deceptive case, the ties just happen to be resolved correctly.
    params3 = np.ones(10, dtype=float)
    assert kendalltau_dist(params3) == 0


def test_rmse_error():
    params1 = np.array([+1.0, -1.2, +0.0, -0.3])
    params2 = params1 - 10.0
    assert np.allclose(rmse(params1, params2), 0)
    params3 = params1 + np.array([+0.1, -0.1, -0.1, +0.1])
    assert np.allclose(rmse(params1, params3), 0.1)


def test_rmse_simple_cases():
    params1 = np.arange(3, dtype=float)
    params2 = np.arange(4, dtype=float)
    with pytest.raises(AssertionError):
        rmse(params1, params2)


def test_log_likelihood_pairwise():
    data1 = ((0, 1),)
    data2 = ((0, 1), (1, 0))
    data3 = ((0, 1), (1, 2), (2, 0))
    params1 = np.zeros(3, dtype=float)
    params2 = np.arange(5, dtype=float)
    assert np.allclose(log_likelihood_pairwise(data1, params1), -log(2))
    assert np.allclose(log_likelihood_pairwise(data2, params1), -2 * log(2))
    assert np.allclose(log_likelihood_pairwise(data3, params1), -3 * log(2))
    assert np.allclose(log_likelihood_pairwise(data1, params2), -log(1 + e))
    assert np.allclose(log_likelihood_pairwise(data2, params2), 1 - 2 * log(1 + e))
    assert np.allclose(
        log_likelihood_pairwise(data3, params2), 3 - log(1 + e) - log(e + e * e) - log(e * e + 1)
    )


def test_log_likelihood_rankings():
    data = ((0, 1, 2, 3), (1, 3, 0))
    params1 = np.zeros(10)
    params2 = np.log(np.linspace(1, 2, num=4))
    assert np.allclose(log_likelihood_rankings(data, params1), -log(4) - 2 * (log(3) + log(2)))
    assert np.allclose(log_likelihood_rankings(data, params2), -5.486092774024455)


def test_log_likelihood_top1():
    data = ((1, (0, 2, 3)), (3, (1, 2)))
    params1 = np.zeros(10)
    params2 = np.log(np.linspace(1, 2, num=4))
    assert np.allclose(log_likelihood_top1(data, params1), -(log(4) + log(3)))
    assert np.allclose(log_likelihood_top1(data, params2), -2.420368128650429)


def test_log_likelihood_network():
    digraph = nx.DiGraph()
    digraph.add_weighted_edges_from(((0, 1, 1.0), (0, 2, 2.0)))
    traffic_in = [0, 2, 4]
    traffic_out = [6, 0, 0]
    params = np.log([1.0, 2.0, 3.0])
    assert np.allclose(
        log_likelihood_network(digraph, traffic_in, traffic_out, params, weight=None),
        2 * log(2 / 5) + 4 * log(3 / 5),
    )
    assert np.allclose(
        log_likelihood_network(digraph, traffic_in, traffic_out, params, weight="weight"),
        2 * log(2 / 8) + 4 * log(3 / 8),
    )


def test_statdist():
    """``statdist`` should return the stationary distribution."""
    gen1 = np.array([[-1, 2 / 3, 1 / 3], [1 / 3, -1, 2 / 3], [2 / 3, 1 / 3, -1]])
    dist1 = np.array([1.0, 1.0, 1.0])
    assert np.allclose(statdist(gen1), dist1)
    gen2 = np.array([[-2 / 3, 2 / 3, 0], [1 / 3, -1, 2 / 3], [0, 1 / 3, -1 / 3]])
    dist2 = np.array([3 / 7, 6 / 7, 12 / 7])
    assert np.allclose(statdist(gen2), dist2)


def test_statdist_single_absorbing_class():
    """
    ``statdist`` should work when the graph is not strongly connected, but has
    a single absorbing class.
    """
    # Markov chain where states 0 and 1 are transient, and 2 and 3 are
    # absorbing. It is weakly but not strongly connected, and has a single
    # absorbing class.
    gen = np.array([[-1, 1, 0, 0], [1, -2, 1, 0], [0, 0, -1, 1], [0, 0, 1, -1]], dtype=float)
    dist = np.array([0.0, 0.0, 2.0, 2.0])
    assert np.allclose(statdist(gen), dist)


def test_statdist_two_absorbing_classes():
    """
    ``statdist`` should fail when the graph is disconnected or has more than
    one absorbing class.
    """
    # Markov chain with two absorbing classes, (0, 1) and (3, 4).
    gen1 = np.array(
        [[-1, 1, 0, 0, 0], [1, -1, 0, 0, 0], [0, 1, -2, 1, 0], [0, 0, 0, -1, 1], [0, 0, 0, 1, -1]],
        dtype=float,
    )
    with pytest.raises(ValueError):
        statdist(gen1)
    # Markov with two disconnected components, (0, 1) and (2, 3).
    gen2 = np.array([[-1, 1, 0, 0], [1, -1, 0, 0], [0, 0, -1, 1], [0, 0, 1, -1]], dtype=float)
    with pytest.raises(ValueError):
        statdist(gen2)


def test_softmax():
    """``softmax`` should work as expected."""
    params1 = np.array([0, 0, 0])
    params2 = np.array([1000, 1000, 2000])
    assert np.allclose(softmax(params1), [1 / 3, 1 / 3, 1 / 3])
    assert np.allclose(softmax(params2), [0, 0, 1])


def test_normal_cdf():
    """``normal_cdf`` should return the value of the normal CDF."""
    for x in 3 * RND.randn(10):
        np.allclose(normal_cdf(x), sps.norm.cdf(x))


def test_normal_pdf():
    """``normal_pdf`` should return the value of the normal PDF."""
    for x in 3 * RND.randn(10):
        np.allclose(normal_pdf(x), sps.norm.pdf(x))


def test_inv_posdef():
    """``inv_posdef`` should return the correct inverse."""
    mat = RND.randn(8, 8)
    mat = mat.dot(mat.T)  # Make it positive semi-definite.
    assert np.allclose(inv_posdef(mat), inv(mat))


def test_generate_params():
    """``generate_params`` should work as expected."""
    params1 = generate_params(10)
    assert len(params1) == 10
    params2 = generate_params(10, ordered=True)
    assert params2.tolist() == sorted(params2)


def test_generate_pairwise():
    """``generate_pairwise`` should work as expected."""
    params = np.exp(RND.rand(10))
    for num in RND.choice(20, size=3, replace=False):
        data = generate_pairwise(params, num)
        assert np.array(data).shape == (num, 2)


def test_generate_rankings():
    """``generate_rankings`` should work as expected."""
    n_items = 10
    params = np.exp(RND.rand(n_items))
    for num in RND.choice(20, size=3, replace=False):
        size = 1 + RND.choice(n_items - 1)
        print(params, num, size)
        data = generate_rankings(params, num, size=size)
        assert np.array(data).shape == (num, size)


def test_compare_choice():
    """``compare`` should work as expected for choices."""
    params1 = np.array([0, 100, -100, -100, -100])
    x1 = compare((3, 0, 2, 4), params1)
    assert x1 == 0
    x2 = compare((3, 0, 1, 4), params1)
    assert x2 == 1
    params2 = np.zeros(10)
    for _ in range(10):
        items = RND.choice(10, size=3, replace=False)
        assert compare(items, params2) in items  # pyright: ignore[reportArgumentType]


def test_compare_rankings():
    """``compare`` should work as expected for rankings."""
    params = np.array([0, 100, -100, -100, -100])
    x1 = compare((3, 0), params, rank=True)
    assert np.array_equal(x1, np.array([0, 3]))
    x2 = compare((3, 0, 1), params, rank=True)
    assert np.array_equal(x2, np.array([1, 0, 3]))


def test_probabilities():
    """``probabilities`` should work as expected."""
    params = np.log([1, 2, 3, 4])
    assert np.allclose(probabilities([0, 2, 3], params), [1 / 8, 3 / 8, 4 / 8])
    assert np.allclose(probabilities([1, 0], params), [2 / 3, 1 / 3])



from unittest.mock import Mock

import numpy as np
import pytest

from choix.lsr import (
    _ilsr,
    ilsr_pairwise,
    ilsr_pairwise_dense,
    ilsr_rankings,
    ilsr_top1,
    lsr_pairwise,
    lsr_pairwise_dense,
    lsr_rankings,
    lsr_top1,
    rank_centrality,
)

from .tutils import iter_testcases

# Tolerance values for calls to `numpy.allclose`.
ATOL = 1e-4
RTOL = 1e-5


def test_ilsr_tolerance():
    """Tolerance affects the number of iterations."""
    vals = [
        np.array([-0.5, 0.5]),
        np.array([-0.3, 0.3]),
        np.array([-0.2, 0.2]),
        np.array([-0.25, 0.25]),
    ]
    lsr = Mock(side_effect=vals)
    est = _ilsr(fun=lsr, params=None, max_iter=100, tol=0.15)
    assert np.array_equal(est, vals[2])
    assert lsr.call_count == 3


def test_ilsr_max_iter():
    """Low `max_iter` raises `RuntimeError`."""
    vals = [
        np.array([-0.5, 0.5]),
        np.array([-0.3, 0.3]),
        np.array([-0.2, 0.2]),
        np.array([-0.25, 0.25]),
    ]
    lsr = Mock(side_effect=vals)
    with pytest.raises(RuntimeError):
        _ilsr(fun=lsr, params=None, max_iter=2, tol=0.01)
    assert lsr.call_count == 2


def test_lsr_pairwise():
    """JSON test cases for LSR pairwise."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["lsr_est"], lsr_pairwise(n_items, data), atol=ATOL, rtol=RTOL)


def test_ilsr_pairwise():
    """JSON test cases for I-LSR pairwise."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["ml_est"], ilsr_pairwise(n_items, data), atol=ATOL, rtol=RTOL)


def test_lsr_pairwise_dense():
    """JSON test cases for LSR pairwise."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        mat = np.zeros((n_items, n_items), dtype=float)
        for a, b in case["data"]:
            mat[a, b] += 1
        assert np.allclose(case["lsr_est"], lsr_pairwise_dense(mat), atol=ATOL, rtol=RTOL)


def test_ilsr_pairwise_dense():
    """JSON test cases for LSR pairwise."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        mat = np.zeros((n_items, n_items), dtype=float)
        for a, b in case["data"]:
            mat[a, b] += 1
        assert np.allclose(case["ml_est"], ilsr_pairwise_dense(mat), atol=ATOL, rtol=RTOL)


def test_rank_centrality():
    """JSON test cases for Rank Centrality."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["rc_est"], rank_centrality(n_items, data), atol=ATOL, rtol=RTOL)


def test_lsr_rankings():
    """JSON test cases for LSR rankings."""
    for case in iter_testcases("rankings"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["lsr_est"], lsr_rankings(n_items, data), atol=ATOL, rtol=RTOL)


def test_ilsr_rankings():
    """JSON test cases for I-LSR rankings."""
    for case in iter_testcases("rankings"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["ml_est"], ilsr_rankings(n_items, data), atol=ATOL, rtol=RTOL)


def test_lsr_top1():
    """JSON test cases for LSR top1."""
    for case in iter_testcases("top1"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["lsr_est"], lsr_top1(n_items, data), atol=ATOL, rtol=RTOL)


def test_ilsr_top1():
    """JSON test cases for I-LSR top1."""
    for case in iter_testcases("top1"):
        n_items = case["n_items"]
        data = case["data"]
        assert np.allclose(case["ml_est"], ilsr_top1(n_items, data), atol=ATOL, rtol=RTOL)



"""(Iterative) Luce Spectral Ranking and related inference algorithms."""

from collections.abc import Callable

import numpy as np
from numpy.typing import NDArray

from .convergence import NormOfDifferenceTest
from .typing import PairwiseData, RankingData, Top1Data
from .utils import exp_transform, log_transform, statdist


def _init_lsr(
    n_items: int,
    alpha: float,
    initial_params: NDArray[np.float64] | None,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Initialize the LSR Markov chain and the weights."""
    if initial_params is None:
        weights = np.ones(n_items)
    else:
        weights = exp_transform(initial_params)
    chain = alpha * np.ones((n_items, n_items), dtype=float)
    return weights, chain


def _ilsr(
    fun: Callable[[NDArray[np.float64] | None], NDArray[np.float64]],
    params: NDArray[np.float64] | None,
    max_iter: int,
    tol: float,
) -> NDArray[np.float64]:
    """Iteratively refine LSR estimates until convergence.

    Raises
    ------
    RuntimeError
        If the algorithm does not converge after ``max_iter`` iterations.
    """
    converged = NormOfDifferenceTest(tol, order=1)
    for _ in range(max_iter):
        params = fun(params)
        if converged(params):
            return params
    raise RuntimeError("Did not converge after {} iterations".format(max_iter))


def lsr_pairwise(
    n_items: int,
    data: PairwiseData,
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
) -> NDArray[np.float64]:
    """Compute the LSR estimate of model parameters.

    This function implements the Luce Spectral Ranking inference algorithm
    [MG15]_ for pairwise-comparison data (see :ref:`data-pairwise`).

    The argument ``initial_params`` can be used to iteratively refine an
    existing parameter estimate (see the implementation of
    :func:`~choix.ilsr_pairwise` for an idea on how this works). If it is set
    to `None` (the default), the all-ones vector is used.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Pairwise-comparison data.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to build the transition rates of the LSR Markov chain.

    Returns
    -------
    params : numpy.ndarray
        An estimate of model parameters.
    """
    weights, chain = _init_lsr(n_items, alpha, initial_params)
    for winner, loser in data:
        chain[loser, winner] += 1 / (weights[winner] + weights[loser])
    chain -= np.diag(chain.sum(axis=1))
    return log_transform(statdist(chain))


def ilsr_pairwise(
    n_items: int,
    data: PairwiseData,
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int = 100,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the ML estimate of model parameters using I-LSR.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given pairwise-comparison data (see :ref:`data-pairwise`), using
    the iterative Luce Spectral Ranking algorithm [MG15]_.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Pairwise-comparison data.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """

    def fun(params: NDArray[np.float64] | None) -> NDArray[np.float64]:
        return lsr_pairwise(n_items=n_items, data=data, alpha=alpha, initial_params=params)

    return _ilsr(fun, initial_params, max_iter, tol)


def lsr_pairwise_dense(
    comp_mat: NDArray[np.float64],
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
) -> NDArray[np.float64]:
    """Compute the LSR estimate of model parameters given dense data.

    This function implements the Luce Spectral Ranking inference algorithm
    [MG15]_ for dense pairwise-comparison data.

    The data is described by a pairwise-comparison matrix ``comp_mat`` such
    that ``comp_mat[i,j]`` contains the number of times that item ``i`` wins
    against item ``j``.

    In comparison to :func:`~choix.lsr_pairwise`, this function is particularly
    efficient for dense pairwise-comparison datasets (i.e., containing many
    comparisons for a large fraction of item pairs).

    The argument ``initial_params`` can be used to iteratively refine an
    existing parameter estimate (see the implementation of
    :func:`~choix.ilsr_pairwise` for an idea on how this works). If it is set
    to `None` (the default), the all-ones vector is used.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    comp_mat : np.array
        2D square matrix describing the pairwise-comparison outcomes.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to build the transition rates of the LSR Markov chain.

    Returns
    -------
    params : np.array
        An estimate of model parameters.
    """
    n_items = comp_mat.shape[0]
    ws, chain = _init_lsr(n_items, alpha, initial_params)
    denom = np.tile(ws, (n_items, 1))
    chain += comp_mat.T / (denom + denom.T)
    chain -= np.diag(chain.sum(axis=1))
    return log_transform(statdist(chain))


def ilsr_pairwise_dense(
    comp_mat: NDArray[np.float64],
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int = 100,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the ML estimate of model parameters given dense data.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given dense pairwise-comparison data.

    The data is described by a pairwise-comparison matrix ``comp_mat`` such
    that ``comp_mat[i,j]`` contains the number of times that item ``i`` wins
    against item ``j``.

    In comparison to :func:`~choix.ilsr_pairwise`, this function is
    particularly efficient for dense pairwise-comparison datasets (i.e.,
    containing many comparisons for a large fraction of item pairs).

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    comp_mat : np.array
        2D square matrix describing the pairwise-comparison outcomes.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """

    def fun(params: NDArray[np.float64] | None) -> NDArray[np.float64]:
        return lsr_pairwise_dense(comp_mat=comp_mat, alpha=alpha, initial_params=params)

    return _ilsr(fun, initial_params, max_iter, tol)


def rank_centrality(n_items: int, data: PairwiseData, alpha: float = 0.0) -> NDArray[np.float64]:
    """Compute the Rank Centrality estimate of model parameters.

    This function implements Negahban et al.'s Rank Centrality algorithm
    [NOS12]_. The algorithm is similar to :func:`~choix.ilsr_pairwise`, but
    considers the *ratio* of wins for each pair (instead of the total count).

    The transition rates of the Rank Centrality Markov chain are initialized
    with ``alpha``. When ``alpha > 0``, this corresponds to a form of
    regularization (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Pairwise-comparison data.
    alpha : float, optional
        Regularization parameter.

    Returns
    -------
    params : numpy.ndarray
        An estimate of model parameters.
    """
    _, chain = _init_lsr(n_items, alpha, None)
    for winner, loser in data:
        chain[loser, winner] += 1.0
    # Transform the counts into ratios.
    idx = chain > 0  # Indices (i,j) of non-zero entries.
    chain[idx] = chain[idx] / (chain + chain.T)[idx]
    # Finalize the Markov chain by adding the self-transition rate.
    chain -= np.diag(chain.sum(axis=1))
    return log_transform(statdist(chain))


def lsr_rankings(
    n_items: int,
    data: RankingData,
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
) -> NDArray[np.float64]:
    """Compute the LSR estimate of model parameters.

    This function implements the Luce Spectral Ranking inference algorithm
    [MG15]_ for ranking data (see :ref:`data-rankings`).

    The argument ``initial_params`` can be used to iteratively refine an
    existing parameter estimate (see the implementation of
    :func:`~choix.ilsr_rankings` for an idea on how this works). If it is set
    to `None` (the default), the all-ones vector is used.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Ranking data.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to build the transition rates of the LSR Markov chain.

    Returns
    -------
    params : numpy.ndarray
        An estimate of model parameters.
    """
    weights, chain = _init_lsr(n_items, alpha, initial_params)
    for ranking in data:
        sum_ = weights.take(ranking).sum()
        for i, winner in enumerate(ranking[:-1]):
            val = 1.0 / sum_
            for loser in ranking[i + 1 :]:
                chain[loser, winner] += val
            sum_ -= weights[winner]
    chain -= np.diag(chain.sum(axis=1))
    return log_transform(statdist(chain))


def ilsr_rankings(
    n_items: int,
    data: RankingData,
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int = 100,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the ML estimate of model parameters using I-LSR.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given ranking data (see :ref:`data-rankings`), using the
    iterative Luce Spectral Ranking algorithm [MG15]_.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Ranking data.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """

    def fun(params: NDArray[np.float64] | None) -> NDArray[np.float64]:
        return lsr_rankings(n_items=n_items, data=data, alpha=alpha, initial_params=params)

    return _ilsr(fun, initial_params, max_iter, tol)


def lsr_top1(
    n_items: int,
    data: Top1Data,
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
) -> NDArray[np.float64]:
    """Compute the LSR estimate of model parameters.

    This function implements the Luce Spectral Ranking inference algorithm
    [MG15]_ for top-1 data (see :ref:`data-top1`).

    The argument ``initial_params`` can be used to iteratively refine an
    existing parameter estimate (see the implementation of
    :func:`~choix.ilsr_top1` for an idea on how this works). If it is set to
    `None` (the default), the all-ones vector is used.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Top-1 data.
    alpha : float
        Regularization parameter.
    initial_params : array_like
        Parameters used to build the transition rates of the LSR Markov chain.

    Returns
    -------
    params : numpy.ndarray
        An estimate of model parameters.
    """
    weights, chain = _init_lsr(n_items, alpha, initial_params)
    for winner, losers in data:
        val = 1 / (weights.take(losers).sum() + weights[winner])
        for loser in losers:
            chain[loser, winner] += val
    chain -= np.diag(chain.sum(axis=1))
    return log_transform(statdist(chain))


def ilsr_top1(
    n_items: int,
    data: Top1Data,
    alpha: float = 0.0,
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int = 100,
    tol: float = 1e-8,
):
    """Compute the ML estimate of model parameters using I-LSR.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given top-1 data (see :ref:`data-top1`), using the
    iterative Luce Spectral Ranking algorithm [MG15]_.

    The transition rates of the LSR Markov chain are initialized with
    ``alpha``. When ``alpha > 0``, this corresponds to a form of regularization
    (see :ref:`regularization` for details).

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Top-1 data.
    alpha : float, optional
        Regularization parameter.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """

    def fun(params: NDArray[np.float64] | None) -> NDArray[np.float64]:
        return lsr_top1(n_items=n_items, data=data, alpha=alpha, initial_params=params)

    return _ilsr(fun, initial_params, max_iter, tol)



from math import log, pi, sqrt

import numpy as np
import pytest

from choix.ep import _ep_pairwise, _log_phi, _match_moments_probit, ep_pairwise

DATA = [(3, 2), (2, 1), (1, 0)]

# Expected results if prior variance = 8.0
LOGIT_TRUE_MEAN = np.array([-2.24876774, -0.54278959, 0.5427896, 2.24876774])
LOGIT_TRUE_COV = np.array(
    [
        [5.13442507, 1.78941794, 0.7363867, 0.33977029],
        [1.78941794, 3.87821895, 1.59597641, 0.73638669],
        [0.7363867, 1.59597641, 3.87821895, 1.78941794],
        [0.33977029, 0.73638669, 1.78941794, 5.13442508],
    ]
)

# Expected results if prior variance = 8.0
PROBIT_TRUE_MEAN = np.array([-2.64311237, -0.71902005, 0.71901978, 2.64311263])
PROBIT_TRUE_COV = np.array(
    [
        [4.40599359, 1.91559034, 1.05514881, 0.62326726],
        [1.91559034, 3.24296482, 1.78629554, 1.0551493],
        [1.05514881, 1.78629554, 3.24296458, 1.91559107],
        [0.62326726, 1.0551493, 1.91559107, 4.40599238],
    ]
)


def test_ep_pairwise_unknown_model():
    """Unknown method should raise a ValueError."""
    with pytest.raises(ValueError):
        ep_pairwise(8, [], 1.0, model="qwerty")  # pyright: ignore[reportArgumentType]


def test_ep_pairwise_logit():
    """EP should work as expected with the logit model."""
    mean, cov = ep_pairwise(4, DATA, 1 / 8)
    assert np.allclose(mean, LOGIT_TRUE_MEAN, atol=1e-3)
    assert np.allclose(cov, LOGIT_TRUE_COV, atol=1e-3)


def test_ep_pairwise_probit():
    """EP should work as expected with the probit model."""
    mean, cov = ep_pairwise(4, DATA, 1 / 8, model="probit")
    assert np.allclose(mean, PROBIT_TRUE_MEAN, atol=1e-3)
    assert np.allclose(cov, PROBIT_TRUE_COV, atol=1e-3)


def test_ep_pairwise_max_iter():
    """Should fail if no convergence after ``max_iter`` iterations."""
    data = [(1, 0), (1, 0)]
    with pytest.raises(RuntimeError):
        _ep_pairwise(2, data, 1.0, _match_moments_probit, max_iter=1, initial_state=None)
    assert _ep_pairwise.iterations == 1  # pyright: ignore[reportFunctionMemberAccess]


def test_ep_pairwise_state():
    """EP should take into account the initial state."""
    data = [(1, 0)]
    tau = np.array([0, 0], dtype=float)
    nu = np.array([0, 0], dtype=float)
    # First run.
    mean1, cov1 = ep_pairwise(2, data, 1.0, initial_state=(tau, nu))
    # Second run, with initial state carried over from first run.
    mean2, cov2 = ep_pairwise(2, data, 1.0, initial_state=(tau, nu))
    assert np.allclose(mean1, mean2)
    assert np.allclose(cov1, cov2)
    assert _ep_pairwise.iterations == 1  # pyright: ignore[reportFunctionMemberAccess]


def test_log_phi():
    """``_log_phi`` should work as expected."""
    # First case: z close to 0.
    res, dres = _log_phi(0)
    assert np.allclose(res, -log(2))
    assert np.allclose(dres, 2 / sqrt(2 * pi))
    # Second case: z small. Ground truth computed with GPy-1.7.7.
    res, dres = _log_phi(-15)
    assert np.allclose(res, -116.13138484571169)
    assert np.allclose(dres, 15.066086827167823)
    # Third case: z positive, large.
    res, dres = _log_phi(100)
    assert np.allclose(res, 0)
    assert np.allclose(dres, 0)



from unittest import mock

import networkx as nx
import numpy as np
import pytest

from choix.mm import (
    choicerank,
    mm_pairwise,
    mm_rankings,
    mm_top1,
)

from .tutils import iter_testcases

RND = np.random.RandomState(42)

# Tolerance values for calls to `numpy.allclose`.
ATOL = 1e-4
RTOL = 1e-5


def test_mm_pairwise():
    """JSON test cases for pairwise data."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        data = case["data"]
        for params in (None, RND.randn(n_items)):
            est = mm_pairwise(n_items, data, initial_params=params)
            assert np.allclose(case["ml_est"], est, atol=ATOL, rtol=RTOL)


def test_mm_rankings():
    """JSON test cases for ranking data."""
    for case in iter_testcases("rankings"):
        n_items = case["n_items"]
        data = case["data"]
        for params in (None, RND.randn(n_items)):
            est = mm_rankings(n_items, data, initial_params=params)
            assert np.allclose(case["ml_est"], est, atol=ATOL, rtol=RTOL)


def test_mm_top1():
    """JSON test cases for top-1 data."""
    for case in iter_testcases("top1"):
        n_items = case["n_items"]
        data = case["data"]
        for params in (None, RND.randn(n_items)):
            est = mm_top1(n_items, data, initial_params=params)
            assert np.allclose(case["ml_est"], est, atol=ATOL, rtol=RTOL)


def test_mm_diverges():
    """Should raise an exception if no convergence after ``max_iter``."""
    # MM performs poorly on cycle datasets.
    data = ((0, 1), (1, 2), (2, 3), (3, 0))
    params = np.array([1.0, 0.1, 0.1, 0.1])
    # Pairwise.
    with pytest.raises(RuntimeError):
        mm_pairwise(4, data, initial_params=params, max_iter=10)
    # Ranking.
    with pytest.raises(RuntimeError):
        mm_rankings(4, data, initial_params=params, max_iter=10)
    # Top-1.
    data = ((0, (1,)), (1, (2,)), (2, (3,)), (3, (0,)))
    with pytest.raises(RuntimeError):
        mm_top1(4, data, initial_params=params, max_iter=10)


def test_choicerank_no_networkx():
    with mock.patch.dict("sys.modules", {"networkx": None}):
        with pytest.raises(ImportError):
            choicerank(None, np.array([]), np.array([]))


def test_choicerank_simple():
    """ChoiceRank should return no-nonsense results on a toy dataset."""
    digraph = nx.DiGraph(incoming_graph_data=[(0, 1), (1, 2), (2, 0)])
    traffic_in = np.array([10, 10, 10])
    traffic_out = np.array([10, 10, 10])
    truth = np.array([0.0, 0.0, 0.0])
    for params in (None, RND.randn(3)):
        est = choicerank(digraph, traffic_in, traffic_out, initial_params=params)
        assert np.allclose(truth, est, atol=ATOL, rtol=RTOL)


def test_choicerank_weighted():
    """ChoiceRank should work with weighted networks."""
    digraph = nx.DiGraph()
    digraph.add_weighted_edges_from(((0, 1, 1.0), (0, 2, 2.0)))
    traffic_in = np.array([0, 10, 20])
    traffic_out = np.array([30, 0, 0])
    truth = np.array([0.0, 0.0, 0.0])
    for params in (None, RND.randn(3)):
        est = choicerank(digraph, traffic_in, traffic_out, initial_params=params, weight="weight")
        assert np.allclose(truth, est, atol=ATOL, rtol=RTOL)


def test_choicerank_complex():
    """Slightly more involved test case for ChoiceRank."""
    # Test data generated using `choicerank.py` from the ChoiceRank's paper
    # supplementary material.
    n_items = 8
    edges = [
        (0, 5),
        (1, 3),
        (1, 4),
        (2, 1),
        (2, 4),
        (2, 6),
        (2, 7),
        (3, 2),
        (3, 6),
        (4, 0),
        (4, 1),
        (5, 2),
        (5, 3),
        (6, 1),
        (6, 5),
        (7, 5),
    ]
    digraph = nx.DiGraph(incoming_graph_data=edges)
    traffic_in = np.array([61, 175, 80, 171, 52, 304, 101, 56])
    traffic_out = np.array([113, 121, 129, 114, 134, 132, 133, 124])
    map_est_1 = [
        0.117729,
        0.29774862,
        -0.72924662,
        0.01879213,
        -0.70499544,
        0.30602036,
        -0.04242447,
        0.73637643,
    ]
    map_est_10 = [
        0.01371515,
        0.18006426,
        -0.5516032,
        0.16807068,
        -0.54010984,
        0.16228541,
        0.01224638,
        0.55533115,
    ]
    for alpha, truth in ((1.0, map_est_1), (10.0, map_est_10)):
        for params in (None, RND.randn(n_items)):
            est = choicerank(digraph, traffic_in, traffic_out, initial_params=params, alpha=alpha)
            assert np.allclose(truth, est, atol=ATOL, rtol=RTOL)



import math
import random
import warnings
from collections.abc import Sequence
from typing import Any, cast

import numpy as np
import scipy.linalg as spl
from numpy.typing import ArrayLike, NDArray
from scipy.linalg import solve_triangular
from scipy.special import logsumexp
from scipy.stats import kendalltau, rankdata

from .typing import PairwiseData, RankingData, Top1Data

SQRT2 = math.sqrt(2.0)
SQRT2PI = math.sqrt(2.0 * math.pi)


def log_transform(weights: ArrayLike) -> NDArray[np.float64]:
    """Transform weights into centered log-scale parameters."""
    params = np.log(weights)
    return params - params.mean()


def exp_transform(params: ArrayLike) -> NDArray[np.float64]:
    """Transform parameters into exp-scale weights."""
    params = np.asarray(params)
    weights = np.exp(params - np.mean(params))
    return (len(weights) / weights.sum()) * weights


def softmax(xs: NDArray[np.float64]) -> NDArray[np.float64]:
    """Stable implementation of the softmax function."""
    ys = xs - np.max(xs)
    exps = np.exp(ys)
    return exps / exps.sum(axis=0)


def normal_cdf(x: float) -> float:
    """Normal cumulative density function."""
    # If X ~ N(0,1), returns P(X < x).
    return math.erfc(-x / SQRT2) / 2.0


def normal_pdf(x: float) -> float:
    """Normal probability density function."""
    return math.exp(-x * x / 2.0) / SQRT2PI


def inv_posdef(mat: NDArray[np.float64]) -> NDArray[np.float64]:
    """Stable inverse of a positive definite matrix."""
    # See:
    # - http://www.seas.ucla.edu/~vandenbe/103/lectures/chol.pdf
    # - http://scicomp.stackexchange.com/questions/3188
    chol = np.linalg.cholesky(mat)
    ident = np.eye(mat.shape[0])
    res = solve_triangular(chol, ident, lower=True, overwrite_b=True)
    return np.transpose(res).dot(res)


def footrule_dist(
    params1: NDArray[np.float64],
    params2: NDArray[np.float64] | None = None,
) -> float:
    r"""Compute Spearman's footrule distance between two models.

    This function computes Spearman's footrule distance between the rankings
    induced by two parameter vectors. Let :math:`\sigma_i` be the rank of item
    ``i`` in the model described by ``params1``, and :math:`\tau_i` be its rank
    in the model described by ``params2``. Spearman's footrule distance is
    defined by

    .. math::

      \sum_{i=1}^N | \sigma_i - \tau_i |

    By convention, items with the lowest parameters are ranked first (i.e.,
    sorted using the natural order).

    If the argument ``params2`` is ``None``, the second model is assumed to
    rank the items by their index: item ``0`` has rank 1, item ``1`` has rank
    2, etc.

    Parameters
    ----------
    params1 : array_like
        Parameters of the first model.
    params2 : array_like, optional
        Parameters of the second model.

    Returns
    -------
    dist : float
        Spearman's footrule distance.
    """
    assert params2 is None or len(params1) == len(params2)
    ranks1 = rankdata(params1, method="average")
    if params2 is None:
        ranks2 = np.arange(1, len(params1) + 1, dtype=float)
    else:
        ranks2 = rankdata(params2, method="average")
    return np.sum(np.abs(ranks1 - ranks2))


def kendalltau_dist(
    params1: NDArray[np.float64],
    params2: NDArray[np.float64] | None = None,
) -> float:
    r"""Compute the Kendall tau distance between two models.

    This function computes the Kendall tau distance between the rankings
    induced by two parameter vectors. Let :math:`\sigma_i` be the rank of item
    ``i`` in the model described by ``params1``, and :math:`\tau_i` be its rank
    in the model described by ``params2``. The Kendall tau distance is defined
    as the number of pairwise disagreements between the two rankings, i.e.,

    .. math::

      \sum_{i=1}^N \sum_{j=1}^N
        \mathbf{1} \{ \sigma_i > \sigma_j \wedge \tau_i < \tau_j \}

    By convention, items with the lowest parameters are ranked first (i.e.,
    sorted using the natural order).

    If the argument ``params2`` is ``None``, the second model is assumed to
    rank the items by their index: item ``0`` has rank 1, item ``1`` has rank
    2, etc.

    If some values are equal within a parameter vector, all items are given a
    distinct rank, corresponding to the order in which the values occur.

    Parameters
    ----------
    params1 : array_like
        Parameters of the first model.
    params2 : array_like, optional
        Parameters of the second model.

    Returns
    -------
    dist : float
        Kendall tau distance.
    """
    assert params2 is None or len(params1) == len(params2)
    ranks1 = rankdata(params1, method="ordinal")
    if params2 is None:
        ranks2 = np.arange(1, len(params1) + 1, dtype=float)
    else:
        ranks2 = rankdata(params2, method="ordinal")
    tau, _ = kendalltau(ranks1, ranks2)
    n_items = len(params1)
    n_pairs = n_items * (n_items - 1) / 2
    return round((n_pairs - n_pairs * cast(float, tau)) / 2)


def rmse(params1: NDArray[np.float64], params2: NDArray[np.float64]) -> float:
    r"""Compute the root-mean-squared error between two models.

    Parameters
    ----------
    params1 : array_like
        Parameters of the first model.
    params2 : array_like
        Parameters of the second model.

    Returns
    -------
    error : float
        Root-mean-squared error.
    """
    assert len(params1) == len(params2)
    params1 = np.asarray(params1) - np.mean(params1)
    params2 = np.asarray(params2) - np.mean(params2)
    sqrt_n = math.sqrt(len(params1))
    return float(np.linalg.norm(params1 - params2, ord=2) / sqrt_n)


def log_likelihood_pairwise(data: PairwiseData, params: NDArray[np.float64]) -> float:
    """Compute the log-likelihood of model parameters."""
    loglik = 0
    for winner, loser in data:
        loglik -= np.logaddexp(0, -(params[winner] - params[loser]))
    return loglik


def log_likelihood_rankings(data: RankingData, params: NDArray[np.float64]) -> float:
    """Compute the log-likelihood of model parameters."""
    loglik: float = 0.0
    params = np.asarray(params)
    for ranking in data:
        for i, winner in enumerate(ranking[:-1]):
            loglik -= logsumexp(params.take(ranking[i:]) - params[winner])  # pyright: ignore[reportOperatorIssue]
    return loglik


def log_likelihood_top1(data: Top1Data, params: NDArray[np.float64]) -> float:
    """Compute the log-likelihood of model parameters."""
    loglik: float = 0.0
    params = np.asarray(params)
    for winner, losers in data:
        idx = np.append(winner, losers)
        loglik -= logsumexp(params.take(idx) - params[winner])  # pyright: ignore[reportOperatorIssue]
    return loglik


def log_likelihood_network(
    digraph: Any,
    traffic_in: Sequence,
    traffic_out: Sequence,
    params: NDArray[np.float64],
    weight: str | None = None,
) -> float:
    """
    Compute the log-likelihood of model parameters.

    If ``weight`` is not ``None``, the log-likelihood is correct only up to a
    constant (independent of the parameters).
    """
    loglik: float = 0.0
    for i in range(len(traffic_in)):
        loglik += traffic_in[i] * params[i]
        if digraph.out_degree(i) > 0:
            neighbors = list(digraph.successors(i))
            if weight is None:
                loglik -= traffic_out[i] * logsumexp(params.take(neighbors))
            else:
                weights = [digraph[i][j][weight] for j in neighbors]
                loglik -= traffic_out[i] * logsumexp(params.take(neighbors), b=weights)
    return loglik


def statdist(generator: ArrayLike) -> NDArray[np.float64]:
    """Compute the stationary distribution of a Markov chain.

    Parameters
    ----------
    generator : array_like
        Infinitesimal generator matrix of the Markov chain.

    Returns
    -------
    dist : numpy.ndarray
        The unnormalized stationary distribution of the Markov chain.

    Raises
    ------
    ValueError
        If the Markov chain does not have a unique stationary distribution.
    """
    generator = np.asarray(generator)
    n = generator.shape[0]
    with warnings.catch_warnings():
        # The LU decomposition raises a warning when the generator matrix is
        # singular (which it, by construction, is!).
        warnings.filterwarnings("ignore")
        lu, piv = spl.lu_factor(generator.T, check_finite=False)
    # The last row contains 0's only.
    left = lu[:-1, :-1]
    right = -lu[:-1, -1]
    # Solves system `left * x = right`. Assumes that `left` is
    # upper-triangular (ignores lower triangle).
    try:
        res = spl.solve_triangular(left, right, check_finite=False)
    except:  # noqa: E722
        # Ideally we would like to catch `spl.LinAlgError` only, but there seems
        # to be a bug in scipy, in the code that raises the LinAlgError (!!).
        raise ValueError(
            "stationary distribution could not be computed. "
            "Perhaps the Markov chain has more than one absorbing class?"
        )
    res = np.append(res, 1.0)
    return (n / res.sum()) * res


def generate_params(
    n_items: int, interval: float = 5.0, ordered: bool = False
) -> NDArray[np.float64]:
    r"""Generate random model parameters.

    This function samples a parameter independently and uniformly for each
    item. ``interval`` defines the width of the uniform distribution.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    interval : float
        Sampling interval.
    ordered : bool, optional
        If true, the parameters are ordered from lowest to highest.

    Returns
    -------
    params : numpy.ndarray
       Model parameters.
    """
    params = np.random.uniform(low=0, high=interval, size=n_items)
    if ordered:
        params.sort()
    return params - params.mean()


def generate_pairwise(params: ArrayLike, n_comparisons: int = 10) -> PairwiseData:
    """Generate pairwise comparisons from a Bradley--Terry model.

    This function samples comparisons pairs independently and uniformly at
    random over the ``len(params)`` choose 2 possibilities, and samples the
    corresponding comparison outcomes from a Bradley--Terry model parametrized
    by ``params``.

    Parameters
    ----------
    params : array_like
        Model parameters.
    n_comparisons : int
        Number of comparisons to be returned.

    Returns
    -------
    data : list of (int, int)
       Pairwise-comparison samples (see :ref:`data-pairwise`).
    """
    params = np.asarray(params)
    n = len(params)
    items = tuple(range(n))
    data = list()
    for _ in range(n_comparisons):
        # Pick the pair uniformly at random.
        a, b = random.sample(items, 2)
        if compare((a, b), params) == a:
            data.append((a, b))
        else:
            data.append((b, a))
    return tuple(data)


def generate_rankings(params: ArrayLike, n_rankings: int, size: int = 3) -> RankingData:
    """Generate rankings according to a Plackett--Luce model.

    This function samples subsets of items (of size ``size``) independently and
    uniformly at random, and samples the correspoding partial ranking from a
    Plackett--Luce model parametrized by ``params``.

    Parameters
    ----------
    params : array_like
        Model parameters.
    n_rankings : int
        Number of rankings to generate.
    size : int, optional
        Number of items to include in each ranking.

    Returns
    -------
    data : list of numpy.ndarray
        A list of (partial) rankings generated according to a Plackett--Luce
        model with the specified model parameters.
    """
    params = np.asarray(params)
    n = len(params)
    items = tuple(range(n))
    data = list()
    for _ in range(n_rankings):
        # Pick the alternatives uniformly at random.
        alts = random.sample(items, size)
        ranking = compare(alts, params, rank=True)
        data.append(ranking)
    return tuple(data)


def compare(items: Sequence[int], params: ArrayLike, rank: bool = False) -> int | NDArray[np.int64]:
    """Generate a comparison outcome that follows Luce's axiom.

    This function samples an outcome for the comparison of a subset of items,
    from a model parametrized by ``params``. If ``rank`` is True, it returns a
    ranking over the items, otherwise it returns a single item.

    Parameters
    ----------
    items : list
        Subset of items to compare.
    params : array_like
        Model parameters.
    rank : bool, optional
        If true, returns a ranking over the items instead of a single item.

    Returns
    -------
    outcome : int or list of int
        The chosen item, or a ranking over ``items``.
    """
    probs = probabilities(items, params)
    if rank:
        return np.random.choice(items, size=len(items), replace=False, p=probs)
    else:
        return np.random.choice(items, p=probs)


def probabilities(items: Sequence[int], params: ArrayLike) -> NDArray[np.float64]:
    """Compute the comparison outcome probabilities given a subset of items.

    This function computes, for each item in ``items``, the probability that it
    would win (i.e., be chosen) in a comparison involving the items, given
    model parameters.

    Parameters
    ----------
    items : list
        Subset of items to compare.
    params : array_like
        Model parameters.

    Returns
    -------
    probs : numpy.ndarray
        A probability distribution over ``items``.
    """
    params = np.asarray(params)
    return softmax(params.take(items))



"""Minorization-maximization inference algorithms."""

import itertools
from collections.abc import Callable
from typing import Any, TypeVar

import numpy as np
from numpy.typing import NDArray

from .convergence import NormOfDifferenceTest
from .typing import PairwiseData, RankingData, Top1Data
from .utils import exp_transform, log_transform

T = TypeVar("T")


def _mm(
    n_items: int,
    data: T,
    initial_params: NDArray[np.float64] | None,
    alpha: float,
    max_iter: int,
    tol: float,
    mm_fun: Callable[
        [int, T, NDArray[np.float64]], tuple[NDArray[np.float64], NDArray[np.float64]]
    ],
) -> NDArray[np.float64]:
    """
    Iteratively refine MM estimates until convergence.

    Raises
    ------
    RuntimeError
        If the algorithm does not converge after `max_iter` iterations.
    """
    if initial_params is None:
        params = np.zeros(n_items)
    else:
        params = initial_params
    converged = NormOfDifferenceTest(tol=tol, order=1)
    for _ in range(max_iter):
        nums, denoms = mm_fun(n_items, data, params)
        params = log_transform((nums + alpha) / (denoms + alpha))
        if converged(params):
            return params
    raise RuntimeError("Did not converge after {} iterations".format(max_iter))


def _mm_pairwise(
    n_items: int,
    data: PairwiseData,
    params: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Inner loop of MM algorithm for pairwise data."""
    weights = exp_transform(params)
    wins = np.zeros(n_items, dtype=float)
    denoms = np.zeros(n_items, dtype=float)
    for winner, loser in data:
        wins[winner] += 1.0
        val = 1.0 / (weights[winner] + weights[loser])
        denoms[winner] += val
        denoms[loser] += val
    return wins, denoms


def mm_pairwise(
    n_items: int,
    data: PairwiseData,
    initial_params: NDArray[np.float64] | None = None,
    alpha: float = 0.0,
    max_iter: int = 10000,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the ML estimate of model parameters using the MM algorithm.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given pairwise-comparison data (see :ref:`data-pairwise`), using
    the minorization-maximization (MM) algorithm [Hun04]_, [CD12]_.

    If ``alpha > 0``, the function returns the maximum a-posteriori (MAP)
    estimate under a (peaked) Dirichlet prior. See :ref:`regularization` for
    details.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Pairwise-comparison data.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    alpha : float, optional
        Regularization parameter.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """
    return _mm(n_items, data, initial_params, alpha, max_iter, tol, _mm_pairwise)


def _mm_rankings(
    n_items: int,
    data: RankingData,
    params: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Inner loop of MM algorithm for ranking data."""
    weights = exp_transform(params)
    wins = np.zeros(n_items, dtype=float)
    denoms = np.zeros(n_items, dtype=float)
    for ranking in data:
        sum_ = weights.take(ranking).sum()
        for i, winner in enumerate(ranking[:-1]):
            wins[winner] += 1
            val = 1.0 / sum_
            for item in ranking[i:]:
                denoms[item] += val
            sum_ -= weights[winner]
    return wins, denoms


def mm_rankings(
    n_items: int,
    data: RankingData,
    initial_params: NDArray[np.float64] | None = None,
    alpha: float = 0.0,
    max_iter: int = 10000,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the ML estimate of model parameters using the MM algorithm.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given ranking data (see :ref:`data-rankings`), using the
    minorization-maximization (MM) algorithm [Hun04]_, [CD12]_.

    If ``alpha > 0``, the function returns the maximum a-posteriori (MAP)
    estimate under a (peaked) Dirichlet prior. See :ref:`regularization` for
    details.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Ranking data.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    alpha : float, optional
        Regularization parameter.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """
    return _mm(n_items, data, initial_params, alpha, max_iter, tol, _mm_rankings)


def _mm_top1(
    n_items: int,
    data: Top1Data,
    params: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Inner loop of MM algorithm for top1 data."""
    weights = exp_transform(params)
    wins = np.zeros(n_items, dtype=float)
    denoms = np.zeros(n_items, dtype=float)
    for winner, losers in data:
        wins[winner] += 1
        val = 1 / (weights.take(losers).sum() + weights[winner])
        for item in itertools.chain([winner], losers):
            denoms[item] += val
    return wins, denoms


def mm_top1(
    n_items: int,
    data: Top1Data,
    initial_params: NDArray[np.float64] | None = None,
    alpha: float = 0.0,
    max_iter: int = 10000,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the ML estimate of model parameters using the MM algorithm.

    This function computes the maximum-likelihood (ML) estimate of model
    parameters given top-1 data (see :ref:`data-top1`), using the
    minorization-maximization (MM) algorithm [Hun04]_, [CD12]_.

    If ``alpha > 0``, the function returns the maximum a-posteriori (MAP)
    estimate under a (peaked) Dirichlet prior. See :ref:`regularization` for
    details.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Top-1 data.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    alpha : float, optional
        Regularization parameter.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The ML estimate of model parameters.
    """
    return _mm(n_items, data, initial_params, alpha, max_iter, tol, _mm_top1)


def _choicerank(
    n_items: int,
    data: tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]],
    params: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Inner loop of ChoiceRank algorithm."""
    weights = exp_transform(params)
    adj, adj_t, traffic_in, traffic_out = data
    # First phase of message passing.
    zs = adj.dot(weights)
    # Second phase of message passing.
    with np.errstate(invalid="ignore"):
        denoms = adj_t.dot(traffic_out / zs)
    return traffic_in, denoms


def choicerank(
    digraph: Any,
    traffic_in: NDArray[np.float64],
    traffic_out: NDArray[np.float64],
    weight: str | None = None,
    initial_params: NDArray[np.float64] | None = None,
    alpha: float = 1.0,
    max_iter: int = 10000,
    tol: float = 1e-8,
) -> NDArray[np.float64]:
    """Compute the MAP estimate of a network choice model's parameters.

    This function computes the maximum-a-posteriori (MAP) estimate of model
    parameters given a network structure and node-level traffic data (see
    :ref:`data-network`), using the ChoiceRank algorithm [MG17]_, [KTVV15]_.

    The nodes are assumed to be labeled using consecutive integers starting
    from 0.

    Parameters
    ----------
    digraph : networkx.DiGraph
        Directed graph representing the network.
    traffic_in : array_like
        Number of arrivals at each node.
    traffic_out : array_like
        Number of departures at each node.
    weight : str, optional
        The edge attribute that holds the numerical value used for the edge
        weight. If None (default) then all edge weights are 1.
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    alpha : float, optional
        Regularization parameter.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Maximum L1-norm of the difference between successive iterates to
        declare convergence.

    Returns
    -------
    params : numpy.ndarray
        The MAP estimate of model parameters.

    Raises
    ------
    ImportError
        If the NetworkX library cannot be imported.
    """
    import networkx as nx

    # Compute the (sparse) adjacency matrix.
    n_items = len(digraph)
    nodes = np.arange(n_items)
    adj = nx.to_scipy_sparse_matrix(digraph, nodelist=nodes, weight=weight)  # pyright: ignore[reportArgumentType]
    adj_t = adj.T.tocsr()
    # Process the data into a standard form.
    traffic_in = np.asarray(traffic_in)
    traffic_out = np.asarray(traffic_out)
    data = (adj, adj_t, traffic_in, traffic_out)
    return _mm(n_items, data, initial_params, alpha, max_iter, tol, _choicerank)



import math
from typing import Literal

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import minimize
from scipy.special import logsumexp

from .typing import PairwiseData, RankingData, Top1Data
from .utils import softmax

MAXEXP = 500


def _safe_exp(x: float) -> float:
    x = max(min(x, MAXEXP), -MAXEXP)
    return math.exp(x)


class PairwiseFcts:
    """Optimization-related methods for pairwise-comparison data.

    This class provides methods to compute the negative log-likelihood (the
    "objective"), its gradient and its Hessian, given model parameters and
    pairwise-comparison data.
    """

    def __init__(self, data: PairwiseData, penalty: float):
        self._data = data
        self._penalty = penalty

    def objective(self, params: NDArray[np.float64]) -> float:
        """Compute the negative penalized log-likelihood."""
        val = self._penalty * np.sum(params**2)
        for win, los in self._data:
            val += np.logaddexp(0, -(params[win] - params[los]))
        return val

    def gradient(self, params: NDArray[np.float64]) -> NDArray[np.float64]:
        grad = 2 * self._penalty * params
        for win, los in self._data:
            z = 1 / (1 + _safe_exp(params[win] - params[los]))
            grad[win] += -z
            grad[los] += +z
        return grad

    def hessian(self, params: NDArray[np.float64]) -> NDArray[np.float64]:
        hess = 2 * self._penalty * np.identity(len(params))
        for win, los in self._data:
            z = _safe_exp(params[win] - params[los])
            val = 1 / (1 / z + 2 + z)
            hess[(win, los), (los, win)] += -val
            hess[(win, los), (win, los)] += +val
        return hess


class Top1Fcts:
    """Optimization-related methods for top-1 data.

    This class provides methods to compute the negative log-likelihood (the
    "objective"), its gradient and its Hessian, given model parameters and
    top-1 data.

    The class also provides an alternative constructor for ranking data.
    """

    def __init__(self, data: Top1Data, penalty: float):
        self._data = data
        self._penalty = penalty

    @classmethod
    def from_rankings(cls, data: RankingData, penalty: float) -> "Top1Fcts":
        """Alternative constructor for ranking data."""
        top1 = list()
        for ranking in data:
            for i, winner in enumerate(ranking[:-1]):
                top1.append((winner, ranking[i + 1 :]))
        return cls(top1, penalty)

    def objective(self, params: NDArray[np.float64]) -> float:
        """Compute the negative penalized log-likelihood."""
        val = self._penalty * np.sum(params**2)
        for winner, losers in self._data:
            idx = np.append(winner, losers)
            val += logsumexp(params.take(idx) - params[winner])  # pyright: ignore[reportOperatorIssue]
        return val

    def gradient(self, params: NDArray[np.float64]) -> NDArray[np.float64]:
        grad = 2 * self._penalty * params
        for winner, losers in self._data:
            idx = np.append(winner, losers)
            zs = softmax(params.take(idx))
            grad[idx] += zs
            grad[winner] += -1
        return grad

    def hessian(self, params: NDArray[np.float64]) -> NDArray[np.float64]:
        hess = 2 * self._penalty * np.identity(len(params))
        for winner, losers in self._data:
            idx = np.append(winner, losers)
            zs = softmax(params.take(idx))
            hess[np.ix_(idx, idx)] += -np.outer(zs, zs)
            hess[idx, idx] += zs
        return hess


def _opt(
    n_items: int,
    fcts: PairwiseFcts | Top1Fcts,
    method: Literal["BFGS", "Newton-CG"],
    initial_params: NDArray[np.float64] | None,
    max_iter: int | None,
    tol: float,
) -> NDArray[np.float64]:
    if initial_params is not None:
        x0 = initial_params
    else:
        x0 = np.zeros(n_items)
    if method == "Newton-CG":
        # `xtol`: Average relative error in solution xopt acceptable for
        # convergence [scipy doc].
        res = minimize(
            fcts.objective,
            x0,
            method=method,
            jac=fcts.gradient,
            hess=fcts.hessian,
            options={"xtol": tol, "maxiter": max_iter},
        )
    elif method == "BFGS":
        # `gtol`: Gradient norm must be less than gtol before successful
        # termination [scipy doc].
        res = minimize(
            fcts.objective,
            x0,
            method=method,
            jac=fcts.gradient,
            options={"gtol": tol, "maxiter": max_iter},
        )
    else:
        raise ValueError("method not known")
    return res.x


def opt_pairwise(
    n_items: int,
    data: PairwiseData,
    alpha: float = 1e-6,
    method: Literal["BFGS", "Newton-CG"] = "Newton-CG",
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int | None = None,
    tol: float = 1e-5,
):
    """Compute the ML estimate of model parameters using ``scipy.optimize``.

    This function computes the maximum-likelihood estimate of model parameters
    given pairwise-comparison data (see :ref:`data-pairwise`), using optimizers
    provided by the ``scipy.optimize`` module.

    If ``alpha > 0``, the function returns the maximum a-posteriori (MAP)
    estimate under an isotropic Gaussian prior with variance ``1 / alpha``. See
    :ref:`regularization` for details.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Pairwise-comparison data.
    alpha : float, optional
        Regularization strength.
    method : str, optional
        Optimization method. Either "BFGS" or "Newton-CG".
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Tolerance for termination (method-specific).

    Returns
    -------
    params : numpy.ndarray
        The (penalized) ML estimate of model parameters.

    Raises
    ------
    ValueError
        If the method is not "BFGS" or "Newton-CG".
    """
    fcts = PairwiseFcts(data, alpha)
    return _opt(n_items, fcts, method, initial_params, max_iter, tol)


def opt_rankings(
    n_items: int,
    data: RankingData,
    alpha: float = 1e-6,
    method: Literal["BFGS", "Newton-CG"] = "Newton-CG",
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int | None = None,
    tol: float = 1e-5,
):
    """Compute the ML estimate of model parameters using ``scipy.optimize``.

    This function computes the maximum-likelihood estimate of model parameters
    given ranking data (see :ref:`data-rankings`), using optimizers provided by
    the ``scipy.optimize`` module.

    If ``alpha > 0``, the function returns the maximum a-posteriori (MAP)
    estimate under an isotropic Gaussian prior with variance ``1 / alpha``. See
    :ref:`regularization` for details.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Ranking data.
    alpha : float, optional
        Regularization strength.
    method : str, optional
        Optimization method. Either "BFGS" or "Newton-CG".
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Tolerance for termination (method-specific).

    Returns
    -------
    params : numpy.ndarray
        The (penalized) ML estimate of model parameters.

    Raises
    ------
    ValueError
        If the method is not "BFGS" or "Newton-CG".
    """
    fcts = Top1Fcts.from_rankings(data, alpha)
    return _opt(n_items, fcts, method, initial_params, max_iter, tol)


def opt_top1(
    n_items: int,
    data: Top1Data,
    alpha: float = 1e-6,
    method: Literal["BFGS", "Newton-CG"] = "Newton-CG",
    initial_params: NDArray[np.float64] | None = None,
    max_iter: int | None = None,
    tol: float = 1e-5,
):
    """Compute the ML estimate of model parameters using ``scipy.optimize``.

    This function computes the maximum-likelihood estimate of model parameters
    given top-1 data (see :ref:`data-top1`), using optimizers provided by the
    ``scipy.optimize`` module.

    If ``alpha > 0``, the function returns the maximum a-posteriori (MAP)
    estimate under an isotropic Gaussian prior with variance ``1 / alpha``. See
    :ref:`regularization` for details.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Top-1 data.
    alpha : float, optional
        Regularization strength.
    method : str, optional
        Optimization method. Either "BFGS" or "Newton-CG".
    initial_params : array_like, optional
        Parameters used to initialize the iterative procedure.
    max_iter : int, optional
        Maximum number of iterations allowed.
    tol : float, optional
        Tolerance for termination (method-specific).

    Returns
    -------
    params : numpy.ndarray
        The (penalized) ML estimate of model parameters.

    Raises
    ------
    ValueError
        If the method is not "BFGS" or "Newton-CG".
    """
    fcts = Top1Fcts(data, alpha)
    return _opt(n_items, fcts, method, initial_params, max_iter, tol)



import functools
from collections.abc import Callable, Sequence
from math import exp, log, pi, sqrt  # Faster than numpy equivalents.
from typing import Literal

import numpy as np
import numpy.random as nprand
from numpy.linalg import norm
from numpy.typing import NDArray
from scipy.special import logsumexp

from .typing import PairwiseData
from .utils import SQRT2, SQRT2PI, inv_posdef, normal_cdf

# EP-related settings.
THRESHOLD = 1e-4

MAT_ONE = np.array([[1.0, -1.0], [-1.0, 1.0]])
MAT_ONE_FLAT = MAT_ONE.ravel()


# Some magic constants for a stable computation of _log_phi(z).
CS = [
    0.00048204,
    -0.00142906,
    0.0013200243174,
    0.0009461589032,
    -0.0045563339802,
    0.00556964649138,
    0.00125993961762116,
    -0.01621575378835404,
    0.02629651521057465,
    -0.001829764677455021,
    2 * (1 - pi / 3),
    (4 - pi) / 3,
    1,
    1,
]
RS = [
    1.2753666447299659525,
    5.019049726784267463450,
    6.1602098531096305441,
    7.409740605964741794425,
    2.9788656263939928886,
]
QS = [
    2.260528520767326969592,
    9.3960340162350541504,
    12.048951927855129036034,
    17.081440747466004316,
    9.608965327192787870698,
    3.3690752069827527677,
]


def ep_pairwise(
    n_items: int,
    data: PairwiseData,
    alpha: float,
    model: Literal["logit", "probit"] = "logit",
    max_iter: int = 100,
    initial_state: tuple | None = None,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Compute a distribution of model parameters using the EP algorithm.

    This function computes an approximate Bayesian posterior probability
    distribution over model parameters, given pairwise-comparison data (see
    :ref:`data-pairwise`). It uses the expectation propagation algorithm, as
    presented, e.g., in [CG05]_.

    The prior distribution is assumed to be isotropic Gaussian with variance
    ``1 / alpha``. The posterior is approximated by a a general multivariate
    Gaussian distribution, described by a mean vector and a covariance matrix.

    Two different observation models are available. ``logit`` (default) assumes
    that pairwise-comparison outcomes follow from a Bradley-Terry model.
    ``probit`` assumes that the outcomes follow from Thurstone's model.

    Parameters
    ----------
    n_items : int
        Number of distinct items.
    data : list of lists
        Pairwise-comparison data.
    alpha : float
        Inverse variance of the (isotropic) prior.
    model : str, optional
        Observation model. Either "logit" or "probit".
    max_iter : int, optional
        Maximum number of iterations allowed.
    initial_state : tuple of array_like, optional
        Natural parameters used to initialize the EP algorithm.

    Returns
    -------
    mean : numpy.ndarray
        The mean vector of the approximate Gaussian posterior.
    cov : numpy.ndarray
        The covariance matrix of the approximate Gaussian posterior.

    Raises
    ------
    ValueError
        If the observation model is not "logit" or "probit".
    """
    if model == "logit":
        match_moments = _match_moments_logit
    elif model == "probit":
        match_moments = _match_moments_probit
    else:
        raise ValueError("unknown model '{}'".format(model))
    return _ep_pairwise(n_items, data, alpha, match_moments, max_iter, initial_state)


def _ep_pairwise(
    n_items: int,
    comparisons: PairwiseData,
    alpha: float,
    match_moments: Callable[[float, float], tuple[float, float, float]],
    max_iter: int,
    initial_state: tuple | None,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Compute a distribution of model parameters using the EP algorithm.

    Raises
    ------
    RuntimeError
        If the algorithm does not converge after ``max_iter`` iterations.
    """
    # Static variable that allows to check the # of iterations after the call.
    _ep_pairwise.iterations = 0  # pyright: ignore[reportFunctionMemberAccess]
    m = len(comparisons)
    prior_inv = alpha * np.eye(n_items)
    if initial_state is None:
        # Initially, mean and covariance come from the prior.
        mean = np.zeros(n_items)
        cov = (1 / alpha) * np.eye(n_items)
        # Initialize the natural params in the function space.
        tau = np.zeros(m)
        nu = np.zeros(m)
        # Initialize the natural params in the space of thetas.
        prec = np.zeros((n_items, n_items))
        xs = np.zeros(n_items)
    else:
        tau, nu = initial_state
        mean, cov, xs, prec = _init_ws(n_items, comparisons, prior_inv, tau, nu)
    for _ in range(max_iter):
        _ep_pairwise.iterations += 1  # pyright: ignore[reportFunctionMemberAccess]
        # Keep a copy of the old parameters for convergence testing.
        tau_old = np.array(tau, copy=True)
        nu_old = np.array(nu, copy=True)
        for i in nprand.permutation(m):
            a, b = comparisons[i]
            # Update mean and variance in function space.
            f_var = cov[a, a] + cov[b, b] - 2 * cov[a, b]
            f_mean = mean[a] - mean[b]
            # Cavity distribution.
            tau_tot = 1.0 / f_var
            nu_tot = tau_tot * f_mean
            tau_cav: float = tau_tot - tau[i]  # pyright: ignore[reportAssignmentType]
            nu_cav: float = nu_tot - nu[i]  # pyright: ignore[reportAssignmentType]
            cov_cav: float = 1.0 / tau_cav
            mean_cav: float = cov_cav * nu_cav
            # Moment matching.
            logpart, dlogpart, d2logpart = match_moments(mean_cav, cov_cav)
            # Update factor params in the function space.
            tau[i] = -d2logpart / (1 + d2logpart / tau_cav)
            delta_tau = tau[i] - tau_old[i]
            nu[i] = (dlogpart - (nu_cav / tau_cav) * d2logpart) / (1 + d2logpart / tau_cav)
            delta_nu = nu[i] - nu_old[i]
            # Update factor params in the weight space.
            prec[(a, a, b, b), (a, b, a, b)] += delta_tau * MAT_ONE_FLAT
            xs[a] += delta_nu
            xs[b] -= delta_nu
            # Update mean and covariance.
            if abs(delta_tau) > 0:
                phi = -1.0 / ((1.0 / delta_tau) + f_var) * MAT_ONE
                upd_mat = cov.take([a, b], axis=0)
                cov = cov + upd_mat.T.dot(phi).dot(upd_mat)
            mean = cov.dot(xs)
        # Recompute the global parameters for stability.
        cov = inv_posdef(prior_inv + prec)
        mean = cov.dot(xs)
        if _converged((tau, nu), (tau_old, nu_old)):
            return mean, cov
    raise RuntimeError("EP did not converge after {} iterations".format(max_iter))


def _log_phi(z: float) -> tuple[float, float]:
    """Stable computation of the log of the Normal CDF and its derivative."""
    # Adapted from the GPML function `logphi.m`.
    if z * z < 0.0492:
        # First case: z close to zero.
        coef = -z / SQRT2PI
        val = functools.reduce(lambda acc, c: coef * (c + acc), CS, 0)
        res = -2 * val - log(2)
        dres = exp(-(z * z) / 2 - res) / SQRT2PI
    elif z < -11.3137:
        # Second case: z very small.
        num = functools.reduce(lambda acc, r: -z * acc / SQRT2 + r, RS, 0.5641895835477550741)
        den = functools.reduce(lambda acc, q: -z * acc / SQRT2 + q, QS, 1.0)
        res = log(num / (2 * den)) - (z * z) / 2
        dres = abs(den / num) * sqrt(2.0 / pi)
    else:
        res = log(normal_cdf(z))
        dres = exp(-(z * z) / 2 - res) / SQRT2PI
    return res, dres


def _match_moments_logit(mean_cav: float, cov_cav: float) -> tuple[float, float, float]:
    # Adapted from the GPML function `likLogistic.m`.
    # First use a scale mixture.
    lambdas = sqrt(2) * np.array([0.44, 0.41, 0.40, 0.39, 0.36])
    cs = np.array(
        [
            1.146480988574439e02,
            -1.508871030070582e03,
            2.676085036831241e03,
            -1.356294962039222e03,
            7.543285642111850e01,
        ]
    )
    arr1, arr2, arr3 = np.zeros(5), np.zeros(5), np.zeros(5)
    for i, x in enumerate(lambdas):
        arr1[i], arr2[i], arr3[i] = _match_moments_probit(x * mean_cav, x * x * cov_cav)
    logpart1: float = logsumexp(arr1, b=cs)  # pyright: ignore[reportAssignmentType]
    dlogpart1 = np.dot(np.exp(arr1) * arr2, cs * lambdas) / np.dot(np.exp(arr1), cs)
    d2logpart1 = (
        np.dot(np.exp(arr1) * (arr2 * arr2 + arr3), cs * lambdas * lambdas)
        / np.dot(np.exp(arr1), cs)
    ) - (dlogpart1 * dlogpart1)
    # Tail decays linearly in the log domain (and not quadratically).
    exponent = -10.0 * (abs(mean_cav) - (196.0 / 200.0) * cov_cav - 4.0)
    if exponent < 500:
        lambd = 1.0 / (1.0 + exp(exponent))
        logpart2 = min(cov_cav / 2.0 - abs(mean_cav), -0.1)
        dlogpart2 = 1.0
        if mean_cav > 0:
            logpart2 = log(1 - exp(logpart2))
            dlogpart2 = 0.0
        d2logpart2 = 0.0
    else:
        lambd, logpart2, dlogpart2, d2logpart2 = 0.0, 0.0, 0.0, 0.0
    logpart = (1 - lambd) * logpart1 + lambd * logpart2
    dlogpart = (1 - lambd) * dlogpart1 + lambd * dlogpart2
    d2logpart = (1 - lambd) * d2logpart1 + lambd * d2logpart2
    return logpart, dlogpart, d2logpart


def _match_moments_probit(mean_cav: float, cov_cav: float) -> tuple[float, float, float]:
    # Adapted from the GPML function `likErf.m`.
    z = mean_cav / sqrt(1 + cov_cav)
    logpart, val = _log_phi(z)
    dlogpart = val / sqrt(1 + cov_cav)  # 1st derivative w.r.t. mean.
    d2logpart = -val * (z + val) / (1 + cov_cav)
    return logpart, dlogpart, d2logpart


def _init_ws(
    n_items: int,
    comparisons: PairwiseData,
    prior_inv: NDArray[np.float64],
    tau: NDArray[np.float64],
    nu: NDArray[np.float64],
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Initialize parameters in the weight space."""
    prec = np.zeros((n_items, n_items))
    xs = np.zeros(n_items)
    for i, (a, b) in enumerate(comparisons):
        prec[(a, a, b, b), (a, b, a, b)] += tau[i] * MAT_ONE_FLAT
        xs[a] += nu[i]
        xs[b] -= nu[i]
    cov = inv_posdef(prior_inv + prec)
    mean = cov.dot(xs)
    return mean, cov, xs, prec


def _converged(
    new: Sequence[NDArray[np.float64]],
    old: Sequence[NDArray[np.float64]],
    threshold: float = THRESHOLD,
) -> bool:
    for param_new, param_old in zip(new, old):
        if norm(param_new - param_old, ord=np.inf) > threshold:
            return False
    return True



import abc
from collections.abc import Callable

import numpy as np
from numpy.typing import NDArray


class ConvergenceTest(metaclass=abc.ABCMeta):
    """Abstract base class for convergence tests.

    Convergence tests should implement a single function, `__call__`, which
    takes a parameter vector and returns a boolean indicating whether or not
    the convergence criterion is met.
    """

    @abc.abstractmethod
    def __call__(self, params: NDArray[np.float64], update: bool = True) -> bool:
        """Test whether convergence criterion is met.

        The parameter `update` controls whether `params` should replace the
        previous parameters (i.e., modify the state of the object).
        """


class NormOfDifferenceTest(ConvergenceTest):
    """Convergence test based on the norm of the difference vector.

    This convergence test computes the difference between two successive
    parameter vectors, and declares convergence when the norm of this
    difference vector (normalized by the number of items) is below `tol`.
    """

    def __init__(self, tol: float = 1e-8, order: int = 1):
        self._tol = tol
        self._ord = order
        self._prev_params = None

    def __call__(self, params: NDArray[np.float64], update: bool = True) -> bool:
        params = np.asarray(params) - np.mean(params)
        if self._prev_params is None:
            if update:
                self._prev_params = params
            return False
        dist = np.linalg.norm(self._prev_params - params, ord=self._ord)
        if update:
            self._prev_params = params
        return bool(dist <= self._tol * len(params))


class ScalarFunctionTest(ConvergenceTest):
    """Convergence test based on a scalar function of the parameters.

    This convergence test computes the values of a scalar function of the
    parameters, and declares convergence when the absolute difference between
    two successive values is below `tol`.

    A typical use case of this class is in conjunction with a log-likelihood
    function.
    """

    def __init__(self, fun: Callable[[NDArray[np.float64]], float], tol: float = 1e-8):
        self._fun = fun
        self._tol = tol
        self._prev_val = None

    def __call__(self, params: NDArray[np.float64], update: bool = True) -> bool:
        val = self._fun(params)
        if self._prev_val is None:
            if update:
                self._prev_val = val
            return False
        dist = abs(val - self._prev_val)
        if update:
            self._prev_val = val
        return dist < self._tol



import glob
import json
import os.path

DATA_ROOT = os.path.join(os.path.dirname(__file__), "data")


def iter_testcases(dtype: str):
    pattern = os.path.join(DATA_ROOT, "testcase-{}-*.json".format(dtype))
    for path in glob.glob(pattern):
        with open(path) as f:
            yield json.load(f)



"""inference algorithms for models based on Luce's choice axiom"""
# Simply import functions defined in modules.

from .ep import ep_pairwise
from .lsr import (
    ilsr_pairwise,
    ilsr_pairwise_dense,
    ilsr_rankings,
    ilsr_top1,
    lsr_pairwise,
    lsr_pairwise_dense,
    lsr_rankings,
    lsr_top1,
    rank_centrality,
)
from .mm import (
    choicerank,
    mm_pairwise,
    mm_rankings,
    mm_top1,
)
from .opt import (
    opt_pairwise,
    opt_rankings,
    opt_top1,
)
from .utils import (
    compare,
    footrule_dist,
    generate_pairwise,
    generate_params,
    generate_rankings,
    kendalltau_dist,
    log_likelihood_network,
    log_likelihood_pairwise,
    log_likelihood_rankings,
    log_likelihood_top1,
    probabilities,
    rmse,
)

__all__ = [
    "ep_pairwise",
    "ilsr_pairwise",
    "ilsr_pairwise_dense",
    "ilsr_rankings",
    "ilsr_top1",
    "lsr_pairwise",
    "lsr_pairwise_dense",
    "lsr_rankings",
    "lsr_top1",
    "rank_centrality",
    "choicerank",
    "mm_pairwise",
    "mm_rankings",
    "mm_top1",
    "opt_pairwise",
    "opt_rankings",
    "opt_top1",
    "compare",
    "footrule_dist",
    "generate_pairwise",
    "generate_params",
    "generate_rankings",
    "kendalltau_dist",
    "log_likelihood_network",
    "log_likelihood_pairwise",
    "log_likelihood_rankings",
    "log_likelihood_top1",
    "probabilities",
    "rmse",
]



import math
from typing import cast

import numpy as np
import pytest
from numpy.typing import NDArray
from scipy.optimize import approx_fprime, check_grad

from choix.opt import (
    PairwiseFcts,
    Top1Fcts,
    opt_pairwise,
    opt_rankings,
    opt_top1,
)

from .tutils import iter_testcases

# `8-random` case.
PAIRWISE_DATA = [
    (7, 3),
    (2, 0),
    (5, 2),
    (4, 2),
    (2, 1),
    (4, 5),
    (6, 3),
    (5, 4),
    (7, 0),
    (2, 3),
    (4, 0),
    (0, 4),
    (6, 5),
    (3, 2),
    (3, 4),
    (3, 4),
    (5, 2),
    (7, 3),
    (7, 6),
    (6, 5),
]
# With regularization set to 0.5.
PAIRWISE_ESTIMATE = [
    -0.51974039,
    -0.46280878,
    -0.31376862,
    -0.06854694,
    -0.3185146,
    -0.00346107,
    0.63032783,
    1.05651257,
]

# Randomly generated case.
TOP1_DATA = [
    (7, (6, 3, 1)),
    (6, (4, 5)),
    (5, (0, 4, 1)),
    (0, (3,)),
    (3, (0, 1)),
]
# With regularization set to 0.5.
TOP1_ESTIMATE = [
    -0.12223311,
    -0.53980682,
    0.0,
    -0.08808004,
    -0.40003309,
    0.24075031,
    0.30131247,
    0.60809028,
]

RND = np.random.RandomState(42)
EPS = math.sqrt(np.finfo(float).eps)

# Tolerance values for calls to `numpy.allclose`.
ATOL = 1e-4
RTOL = 1e-5


def _test_gradient(n_items: int, fcts: PairwiseFcts | Top1Fcts):
    """Helper for testing the gradient of objective functions."""
    for sigma in np.linspace(1, 20, num=10):
        xs = sigma * RND.randn(n_items)
        val = cast(NDArray, approx_fprime(xs, fcts.objective, EPS))
        err = check_grad(fcts.objective, fcts.gradient, xs, epsilon=EPS)
        assert abs(err / np.linalg.norm(val)) < 1e-5


def _test_hessian(n_items: int, fcts: PairwiseFcts | Top1Fcts):
    """Helper for testing the hessian of objective functions."""
    for sigma in np.linspace(1, 20, num=10):
        xs = sigma * RND.randn(n_items)
        for i in range(n_items):

            def obj(xs: NDArray[np.float64]) -> float:
                return fcts.gradient(xs)[i]

            def grad(xs: NDArray[np.float64]) -> NDArray[np.float64]:
                return fcts.hessian(xs)[i]

            val = cast(NDArray, approx_fprime(xs, obj, EPS))
            err = check_grad(obj, grad, xs, epsilon=EPS)
            assert abs(err / np.linalg.norm(val)) < 1e-5


def test_pairwise_gradient():
    """Gradient of pairwise-data objective should be correct."""
    fcts = PairwiseFcts(PAIRWISE_DATA, 0.2)
    _test_gradient(8, fcts)


def test_pairwise_hessian():
    """Hessian of pairwise-data objective should be correct."""
    fcts = PairwiseFcts(PAIRWISE_DATA, 0.2)
    _test_hessian(8, fcts)


def test_top1_gradient():
    """Gradient of top1-data objective should be correct."""
    fcts = Top1Fcts(TOP1_DATA, 0.2)
    _test_gradient(8, fcts)


def test_top1_hessian():
    """Hessian of top1-data objective should be correct."""
    fcts = Top1Fcts(TOP1_DATA, 0.2)
    _test_hessian(8, fcts)


def test_opt_pairwise_simple():
    """Simple test where regularization is needed (ML does not exist)."""
    for method in ("BFGS", "Newton-CG"):
        for params in (None, RND.randn(8)):
            est = opt_pairwise(8, PAIRWISE_DATA, alpha=0.5, method=method, initial_params=params)
            assert np.allclose(est, PAIRWISE_ESTIMATE, atol=ATOL, rtol=RTOL)


def test_opt_top1_simple():
    """Simple test where regularization is needed (ML does not exist)."""
    for method in ("BFGS", "Newton-CG"):
        for params in (None, RND.randn(8)):
            est = opt_top1(8, TOP1_DATA, alpha=0.5, method=method, initial_params=params)
            assert np.allclose(est, TOP1_ESTIMATE, atol=ATOL, rtol=RTOL)


def test_opt_pairwise_extreme():
    """MAP estimate takes extreme values (almost divergent)."""
    data = ((0, 1), (1, 2))
    for method in ("BFGS", "Newton-CG"):
        est = opt_pairwise(3, data, alpha=0.00001, method=method)
        assert est[0] > est[1] > est[2]


def test_opt_top1_extreme():
    """MAP estimate takes extreme values (almost divergent)."""
    data = ((0, (1,)), (1, (2,)))
    for method in ("BFGS", "Newton-CG"):
        est = opt_top1(3, data, alpha=0.00001, method=method)
        assert est[0] > est[1] > est[2]


def test_opt_unknown_method():
    """Unknown method should raise a ValueError."""
    with pytest.raises(ValueError):
        opt_pairwise(8, [], method="qwerty")  # pyright: ignore[reportArgumentType]
    with pytest.raises(ValueError):
        opt_rankings(8, [], method="qwerty")  # pyright: ignore[reportArgumentType]
    with pytest.raises(ValueError):
        opt_top1(8, [], method="qwerty")  # pyright: ignore[reportArgumentType]


def test_opt_pairwise_json():
    """JSON test cases for pairwise-data ML estimator."""
    for case in iter_testcases("pairwise"):
        n_items = case["n_items"]
        data = case["data"]
        for method in ("BFGS", "Newton-CG"):
            est = opt_pairwise(n_items, data, alpha=0.0, method=method)
            assert np.allclose(case["ml_est"], est, atol=ATOL, rtol=RTOL)


def test_opt_rankings_json():
    """JSON test cases for ranking-data ML estimator."""
    for case in iter_testcases("rankings"):
        n_items = case["n_items"]
        data = case["data"]
        for method in ("BFGS", "Newton-CG"):
            est = opt_rankings(n_items, data, alpha=0.0, method=method)
            assert np.allclose(case["ml_est"], est, atol=ATOL, rtol=RTOL)


def test_opt_top1_json():
    """JSON test cases for top1-data ML estimator."""
    for case in iter_testcases("top1"):
        n_items = case["n_items"]
        data = case["data"]
        for method in ("BFGS", "Newton-CG"):
            est = opt_top1(n_items, data, alpha=0.0, method=method)
            assert np.allclose(case["ml_est"], est, atol=ATOL, rtol=RTOL)
